/**
  Implements a priority queue which allows the following operations, 
  each one in a in O(log n) time.
  - inserting an item 
  - getting the minimum key item
  - deleting the minimum key item
  - merging two queues
  It has been implemented by alexandre-mesle.com for java 1.5
*/

public class BinomialHeap<T extends Comparable<T>>
{

    /************************************************/

    private LinkedList<BinomialTree<T>> binomialTrees;
    
    /************************************************/

    /**
       Creates an empty queue.
     */

    public BinomialHeap()
    {
	binomialTrees = null;
    }
    
    /************************************************/

    /**
       Creates a one item queue.
     */

    public BinomialHeap(T data)
    {
	binomialTrees = new LinkedList<BinomialTree<T>> 
	    (new BinomialTree<T>(data));  
    }
    
    /************************************************/

    BinomialHeap(LinkedList<BinomialTree<T>> l)    
    {
	binomialTrees = l;
    }

    /************************************************/

    /**
       Returns trus if and only if the queue doesn't contain
       any item.
     */

    public boolean isEmpty()
    {
	return binomialTrees == null;
    }
    
    /************************************************/

    private static <K extends Comparable<K>> LinkedList<BinomialTree<K>> 
	findMinLink(LinkedList<BinomialTree<K>> link, 
		    LinkedList<BinomialTree<K>> acc)  
    {
	if (link == null)
	    return acc;
	K linkData = link.getData().getData();
	K accData = acc.getData().getData();
	if (linkData.compareTo(accData) < 1)
	    return findMinLink(link.getNext(), link);
	return findMinLink(link.getNext(), acc);
    }

    /************************************************/

    /**
       Returns the item wich has the minimum key.
     */

    public T getMin()
    {
	LinkedList<BinomialTree<T>> minNode = findMinLink(binomialTrees, 
							  binomialTrees);
	if (minNode == null)
	    return null;
	return minNode.getData().getData(); 
    }
    
    
    /************************************************/

    /**
       Deletes the minimum key item.
     */

    public void deleteMin()
    {
	LinkedList<BinomialTree<T>> minLink = 
	    findMinLink(binomialTrees, binomialTrees);
	binomialTrees = binomialTrees.deleteLink(minLink);
	BinomialHeap<T> sons = minLink.getData().extractSons(); 
	fusion(sons);
    }
    
    /************************************************/

    /**
       Deletes the minimum key item.
     */

    public void insert(T c)
    {
	fusion(new BinomialHeap<T>(c));
    }
    
    /************************************************/

    /**
       Adds to the current queue all the item of the queue b. 
       Use it only if b doesn't have to used again.
     */

    public void fusion(BinomialHeap<T> b)
    {
	binomialTrees = 
	    clean(merge(binomialTrees, b.binomialTrees));
    }
    
    /************************************************/

    private static <K extends Comparable<K>> LinkedList<BinomialTree<K>> 
	merge(LinkedList<BinomialTree<K>> l1, 
	      LinkedList<BinomialTree<K>> l2) 
    {
	if (l1 == null)
	    return l2;
	if (l2 == null)
	    return l1;	
	if (l1.getData().compareTo(l2.getData()) <= 0)
	    {
		l1.setNext(merge(l1.getNext(), l2));
		return l1;
	    }
	else
	    return merge(l2, l1);
    }
    
    /************************************************/

    private static <K extends Comparable<K>> LinkedList<BinomialTree<K>> 
	clean(LinkedList<BinomialTree<K>> l)
    {
	if (l == null)
	    return null;
	if (l.getNext() == null)
	    return l;
	BinomialTree<K> first = l.getData();
	BinomialTree<K> second = l.getNext().getData();
	if (first.getOrder() != second.getOrder())
	    {
		l.setNext(clean(l.getNext()));
		return l;
	    }
	if (l.getNext().getNext() == null)
	    return new LinkedList<BinomialTree<K>>
		(first.fusion(second), 
		 l.getNext().getNext()); 
	BinomialTree<K> third = l.getNext().getNext().getData();
	if (first.getOrder() == third.getOrder())
	    {
		l.setNext(clean(l.getNext()));
		return l;
	    }
	else
	    return new LinkedList<BinomialTree<K>>
		(first.fusion(second), 
		 l.getNext().getNext()); 
    }
    
    /************************************************/

    /*

    public static void main(String[] args)
    {
	BinomialHeap<Integer> h = new BinomialHeap<Integer>();
	for (int i = 1 ; i < 3000 ; i++)
	    h.insert(new Integer((i*i*i*i*i*i*i) % 127));
	while(!h.isEmpty())
	    {
		System.out.print(h.getMin() + " -> ");
		h.deleteMin();
	    }
    }
   
    */

}